{-----------------------------------------------------------------------------
 Unit Name: mustapas
 Author:    g.desanctis
 Date:      25-mag-2020
 Purpose:   Mustache template engine, featuring:
     - customizable start and end tags (always of 2 chars)
     - variables as context object member (example: obj.varname)
     - array sections
     - conditional sections
     - nested sections
     - comments
 History:
 2016 - initial version for Freepascal by Mario Ray Mahardhika
 2020 - extended version for Delphi by Giandomenico De Sanctis

-----------------------------------------------------------------------------
 The MIT License (MIT)

Copyright (c) 2016 Mario Ray Mahardhika, 2020 Giandomenico De Sanctis

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

 -----------------------------------------------------------------------------}

unit mustapas;


interface

uses
  Classes,SysUtils,
  superobject;

type

  TCacheGetFunction = function (const Template: TStream; const Context: ISuperObject; Result: TStream): Boolean;
  TCachePutFunction = procedure (const Template: TStream; const Context: ISuperObject; const Result: TStream);

  ETokenize = class(Exception) end;
  ERender = class(Exception) end;

 type
   TRenderEngine = class
   private
      templStream: TStringStream;
      resStream:   TStringStream;

      fStartTag: string[2];
      fEndTag:   string[2];

      startTagChanged: Boolean;
      endTagChanged: Boolean;
   public
    constructor Create();
    destructor Destroy(); override;
    procedure setDelimiterTags(const startTag: string; const endTag: string);
    procedure renderTemplate(const Template: string; const Context: ISuperObject; var Result: string);
   end;

var
  OnGetCache: TCacheGetFunction;
  OnPutCache: TCachePutFunction;

function Escape(const S: String): String; inline;
procedure Render(const Template: TStream; const Context: ISuperObject; Result: TStream);

implementation

uses
  StrUtils, streamIO;

const
  START_TAG_1  = '<';
  START_TAG_2  = '%';

  END_TAG_1    = '%';
  END_TAG_2    = '>';

  START_TAG    = START_TAG_1+START_TAG_2;
  END_TAG      = END_TAG_1  +END_TAG_2;

function Escape(const S: String): String;
begin
  Result := StringReplace(S,''#13,'\r',[rfReplaceAll]);
  Result := StringReplace(Result,''#10,'\n',[rfReplaceAll]);
end;

{ token }

type
  TTokenKind = (
    tkText,tkEscapedVariable,tkUnescapedVariable,tkOpenSection,tkCloseSection,
    tkComment
  );

  PToken = ^TToken;
  TToken = record
    Line,Col: LongWord;
    Kind: TTokenKind;
    Lexeme: String;
    ContextBasePath: String;
  end;

  TTokenList = array of TToken;

  TSection = record
    sectionPath     : string;
    sectionContainer: string;
    sectionNode: ISuperObject;
    sectionTokens: TTokenList;
    sectionContext: ISuperObject;
  end;


procedure AddToken(const ALine,ACol: LongWord; const AKind: TTokenKind;
  const ALexeme,AContextBasePath: String; var Tokens: TTokenList);
var
  n: Integer;
  t: TToken;
begin
  // WriteLn('Token ',AKind,' of "',Escape(ALexeme),'" has been added');
  with t do begin
    Line := ALine;
    Col := ACol;
    Kind := AKind;
    Lexeme := ALexeme;
    ContextBasePath := AContextBasePath;
  end;
  n := Length(Tokens);
  SetLength(Tokens,n + 1);
  Tokens[n] := t;
end;

{ lexer }
function RPos(SubStr: string; S: string): integer;
var
  I: integer;
begin
  SubStr := ReverseString(SubStr);
  S := ReverseString(S);
  I := pos(SubStr, S);
  if I <> 0 then
    I := (Length(S) + 1) - (I + Length(SubStr) - 1);
  Result := I;
end;

const
  EOFChar     = #26;
  LFChar      = #10;

type
  TLexerState = (
    lsText,lsPossiblyTagStart,lsTagStart,lsPossiblyTagEnd,lsTagEnd,
    lsEscapedVariable,lsUnescapedVariable,lsPossiblyUnescapedVariableEnd,lsUnescapedVariableAmp,
    lsOpenSection,lsCloseSection, lsComment,
    lsEOF
  );

var
  Input: Text;
  LexerState: TLexerState;
  Line,Col: LongWord; // current lexer position
  LineBuf: String;    // line buffer, holds current line contents of the source file

function ReadChar: Char; // read next character from the input stream
begin
  if EOF(Input) then begin
    Result := EOFChar;
  end else begin
    Read(Input,Result);
    if Result = LFChar then begin
      Inc(Line);
      Col := 0;
    end else begin
      Inc(Col);
    end;
  end;
end;

function Tokenize: TTokenList;
var
  StartLine,StartCol: LongWord;
  Kind: TTokenKind;
  Lexeme,ContextBasePath: String;
  Look: Char;
  TempState: TLexerState;

  procedure InitFields;
  begin
    StartLine := Line;
    StartCol := Col;
    Lexeme := EmptyStr;
  end;

  procedure tokenError();
  begin
    raise ETokenize.Create(
      Format('Invalid tag char: %s lexeme: %s start line: %d start col: %d',
       [Look, lexeme, startline, startcol]) );
  end;

 procedure HandleText;
  begin
    case Look of
//      '{': begin
      START_TAG_1: begin
          LexerState := lsPossiblyTagStart;
      end;
      else begin
        Lexeme := Lexeme + Look;
      end;
    end;
    Look := ReadChar;
  end;

  procedure HandlePossiblyTagStart;
  begin
    case Look of
//      '{': begin
      START_TAG_2: begin
        if Lexeme <> EmptyStr then begin
          AddToken(StartLine,StartCol,Kind,Lexeme,ContextBasePath,Result);
          InitFields;
        end;
        LexerState := lsTagStart;
      end;
      else begin
        Lexeme := Lexeme + START_TAG_1 + Look;
        LexerState := lsText;
      end;
    end;
    Look := ReadChar;
  end;

  procedure HandleTagStart;
  begin
    case Look of
      ' ': begin
        //ignore blanks after start tag
      end;
      '{': begin
        LexerState := lsUnescapedVariable;
        Kind := tkUnescapedVariable;
      end;
      '&': begin
        LexerState := lsUnescapedVariableAmp;
        Kind := tkUnescapedVariable;
      end;
      '#': begin
        LexerState := lsOpenSection;
        Kind := tkOpenSection;
      end;
      '/': begin
        LexerState := lsCloseSection;
        Kind := tkCloseSection;
      end;
      '!': begin
        LexerState := lsComment;
        Kind := tkComment;
      end;
      'a'..'z','A'..'Z','_', '$': begin
        LexerState := lsEscapedVariable;
        Kind := tkEscapedVariable;
        Lexeme := Look;
      end;
      else begin
        tokenError;
      end;
    end;
    Look := ReadChar;
  end;

  procedure HandlePossiblyTagEnd;
  var
    p: integer;
  begin
    case Look of
//      '}': begin
      END_TAG_2: begin
        Lexeme := Trim(Lexeme);
        case TempState of
          lsOpenSection: begin
              ContextBasePath := ContextBasePath + Lexeme + '.';
              Lexeme:='';
              end;
          lsCloseSection: begin
            if ContextBasePath = EmptyStr then
              raise Exception.Create('Closing unopened section: ' + Lexeme)
            else begin
              // delete '.' as last character
              Delete(ContextBasePath,length(ContextBasePath),1);
              p := RPos('.',ContextBasePath);
              if p>0 then
                 Delete(ContextBasePath,p+1,Length(ContextBasePath) - p )
              else
                 ContextBasePath:='';
            end;
          end;
        end;
        if Kind<>tkComment then
        begin
          AddToken(StartLine,StartCol,Kind,Lexeme,ContextBasePath,Result);
        end;
        InitFields;
        LexerState := lsTagEnd;
      end;
      else begin
//        Lexeme := Lexeme + '}' + Look;
        Lexeme := Lexeme + END_TAG_1 + Look;
        LexerState := TempState;
      end;
    end;
    Look := ReadChar;
  end;

  procedure HandleTagEnd;
  begin
    LexerState := lsText;
    Kind := tkText;
  end;

  procedure HandleVariable;
  begin
    case Look of
//      '}': begin
      END_TAG_1: begin
        TempState := LexerState;
        if LexerState = lsEscapedVariable then
          LexerState := lsPossiblyTagEnd
        else
          LexerState := lsPossiblyUnescapedVariableEnd;
      end;
      'a'..'z','A'..'Z','0'..'9','_',' ','.','$': begin
        Lexeme := Lexeme + Look;
      end;
      else begin
        tokenError;
      end;
    end;
    Look := ReadChar;
  end;

  procedure HandlePossiblyUnescapedVariableEnd;
  begin
    case Look of
//      '}': begin
      END_TAG_1: begin
        LexerState := lsPossiblyTagEnd;
      end;
      else begin
        tokenError;
      end;
    end;
    Look := ReadChar;
  end;

  procedure HandleUnescapedVariableAmp;
  begin
    case Look of
//      '}': begin
      END_TAG_1: begin
        TempState := LexerState;
        LexerState := lsPossiblyTagEnd;
      end;
      'a'..'z','A'..'Z','0'..'9','_',' ','.': begin
        Lexeme := Lexeme + Look;
      end;
      else begin
        tokenError;
      end;
    end;
    Look := ReadChar;
  end;

  procedure HandleSection;
  begin
    case Look of
//      '}': begin
      END_TAG_1: begin
        TempState := LexerState;
        LexerState := lsPossiblyTagEnd;
      end;
      'a'..'z','A'..'Z','0'..'9','_',' ','.', '$': begin
        Lexeme := Lexeme + Look;
      end;
      else begin
        tokenError;
      end;
    end;
    Look := ReadChar;
  end;

  procedure HandleComment;
  begin
    case Look of
      END_TAG_1: begin
        TempState := LexerState;
        LexerState := lsPossiblyTagEnd;
      end;
      else begin
        Lexeme := Lexeme + Look;
      end;
    end;
    Look := ReadChar;
  end;

begin
  Initialize(Result);
  LexerState := lsText;

  InitFields;
  ContextBasePath := EmptyStr;
  Look := ReadChar;
  Kind := tkText;
  while Look <> EOFChar do begin
    case LexerState of
      lsText                        : HandleText;
      lsPossiblyTagStart            : HandlePossiblyTagStart;
      lsTagStart                    : HandleTagStart;
      lsPossiblyTagEnd              : HandlePossiblyTagEnd;
      lsTagEnd                      : HandleTagEnd;
      lsEscapedVariable,
      lsUnescapedVariable           : HandleVariable;
      lsPossiblyUnescapedVariableEnd: HandlePossiblyUnescapedVariableEnd;
      lsUnescapedVariableAmp        : HandleUnescapedVariableAmp;
      lsOpenSection,lsCloseSection  : HandleSection;
      lsComment                     : HandleComment;
    end;
  end;
  if Lexeme <> EmptyStr then begin
    AddToken(StartLine,StartCol,Kind,Lexeme,ContextBasePath,Result);
  end;
end;

function HTMLEncode(const Data: string): string;
var
  iPos, i: Integer;

  procedure Encode(const AStr: String);
  begin
    Move(AStr[1], result[iPos], Length(AStr) * SizeOf(Char));
    Inc(iPos, Length(AStr));
  end;

begin
  SetLength(result, Length(Data) * 6);
  iPos := 1;
  for i := 1 to length(Data) do
    case Data[i] of
      '<': Encode('&lt;');
      '>': Encode('&gt;');
      '&': Encode('&amp;');
      '"': Encode('&quot;');
    else
      result[iPos] := Data[i];
      Inc(iPos);
    end;
  SetLength(result, iPos - 1);
end;



procedure InterpretRec(const Tokens: TTokenList; const Context: ISuperObject; var Output: TextFile);
var
  Token: TToken;
  Value,BasePath: String;
  Node: ISuperObject;
  curSection: TSection;
  p: integer;
  openSection: Boolean;

  procedure addSecToken();
  var
    n: Integer;
  begin
    n := Length(curSection.sectionTokens);
    SetLength(curSection.sectionTokens,n + 1);
    curSection.sectionTokens[n] := Token;
  end;

  procedure findNode();
  var
    l: Integer;
  begin
    BasePath := Token.ContextBasePath;
    if Token.Lexeme='' then
        Delete(BasePath, RPos('.', BasePath), 1);

    for l:=1 to 10 do
    begin
      Node :=Context[BasePath+Token.Lexeme];
      if Assigned(Node) or (BasePath = EmptyStr) then Break;
      // delete final '.'
      if RPos('.', BasePath)=Length(BasePath) then
        Delete(BasePath, Length(BasePath), 1);
      p := RPos('.',BasePath);
      if p > 0 then Delete(BasePath,p+1,Length(BasePath) - p )
      else
        BasePath:='';
    end;
  end;


  procedure renderField();
  begin
        if Node.IsType(TSuperType.stInt) then
            Value := IntToStr(Node.AsInteger);

        if Node.IsType(TSuperType.stDouble) then
        begin
            Value := FloatToStr(Node.AsDouble);
        end;
        //jtString:
        if Node.IsType(TSuperType.stString) then
            if Token.Kind = tkEscapedVariable then
              Value := HTMLEncode(Node.AsString)
            else
              Value := Node.AsString;
        //jtBoolean:
        if Node.IsType(TSuperType.stBoolean) then
          Value := BoolToStr(Node.AsBoolean);
        //jtNull: ;
        // no action
        //jtArray:
        if Node.IsType(TSuperType.stArray) then
          raise Exception.Create('Can''t interpolate array directly');
        //jtObject:
        if Node.IsType(TSuperType.stObject) then
          raise Exception.Create('Can''t interpolate object directly');
        //jtUnknown:
        //  raise Exception.Create('Don''t know how to interpret "' + Node.AsJSON + '"');

  end;


  procedure processSection();
  var
    i: Integer;
    ne: Integer;
    sectionContext: ISuperObject;
  begin
     if (curSection.sectionNode.IsType(TSuperType.stBoolean))
        and (curSection.sectionNode.AsBoolean=True) then
     begin
        sectionContext:=Context;
        InterpretRec(curSection.sectionTokens, sectionContext, Output);
     end;
     if curSection.sectionNode.IsType(TSuperType.stArray) then
     begin
         ne:= curSection.sectionNode.AsArray.Length;
         for I:=0 to ne-1 do
         begin
            Node:=ISuperObject(curSection.sectionNode.AsArray[i]);
            sectionContext:=Context;
            sectionContext.O[curSection.sectionContainer]:= Node;
            InterpretRec(curSection.sectionTokens, sectionContext, Output);
         end;
     end;
  end;

  procedure readSection();
  begin
      if (Token.Kind = tkCloseSection)
         and (Token.ContextBasePath+Token.Lexeme= curSection.sectionContainer) then
      begin
            if curSection.sectionNode<>nil then
                processSection;
            curSection.sectionNode:=nil;
            curSection.sectionContainer:='';
            openSection:=False;
      end else
        addSecToken;
  end;


begin
  openSection:=False;
  for Token in Tokens do
    if (openSection) then
    begin
      readSection;
    end else
    begin
        case Token.Kind of
          tkText: begin
                Write(Output,Token.Lexeme);
            end;
          tkEscapedVariable,tkUnescapedVariable: begin
            Value := '';
              findNode;
              if Node<>nil then
              begin
                 renderField;
              end;
              Write(Output,Value);
          end;
          tkOpenSection: begin
            openSection:=True;
            curSection.sectionContainer:=Token.ContextBasePath;
            with curSection do
            begin
              if sectionContainer[Length(sectionContainer)]='.' then
                  Delete(sectionContainer,Length(sectionContainer),1);
            end;
            SetLength(curSection.sectionTokens, 0);
            findNode;
            if Node<>nil then
            begin
              curSection.sectionNode:=Node;
            end;
          end;
        end;
    end;
end;

procedure Interpret(const Tokens: TTokenList; const Context: ISuperObject; Result: TStream);
var
  Output: TextFile;
begin
  AssignStream(Output,Result);
  Rewrite(Output);
  InterpretRec(Tokens, Context, output);
  CloseFile(output);
end;

procedure Render(const Template: TStream; const Context: ISuperObject; Result: TStream);
var
  Tokens: TTokenList;
  Token: TToken;
begin
  if not Assigned(OnGetCache) or not OnGetCache(Template,Context,Result) then begin
    // initialize input and lexer
    AssignStream(Input,Template);
    Reset(Input);
    Line := 1;
    Col := 0;
    Tokens := Tokenize;
    Interpret(Tokens,Context,Result);
    if Assigned(OnPutCache) then OnPutCache(Template,Context,Result);
  end;
end;

{**************** TRenderEngine  ************************************}
constructor TRenderEngine.Create;
begin
  inherited Create;
   templStream:=TStringStream.Create;
   resStream:=TStringStream.Create;
   fStartTag:=START_TAG;
   fEndTag:=END_TAG;
   startTagChanged:=False;
   endTagChanged:=False;
end;

destructor TRenderEngine.Destroy;
begin
   templStream.Free;
   resStream.Free;
  inherited Destroy;
end;

procedure TRenderEngine.renderTemplate(const Template: string;
            const Context: ISuperObject; var Result: string);
var
  temp: string;
begin
   templStream.Clear;
   if (startTagChanged) or (endTagChanged) then
   begin
     if startTagChanged then
     begin
       // a blank after tag reduce substitution errors, especially with double '{{' that
       // can be confused with Javascript parenthesis
       temp:=StringReplace(Template, fStartTag+' ', START_TAG, [rfReplaceAll] );
     end;
     if endTagChanged then
     begin
       temp:=StringReplace(temp, ' '+fEndTag, END_TAG, [rfReplaceAll] );
     end;
     templStream.WriteString(temp);
   end
    else
   begin
      templStream.WriteString(Template);
   end;
   resStream.Clear;
   render(templStream, Context, resStream);
   Result:=resStream.DataString;
end;


procedure TRenderEngine.setDelimiterTags(const startTag: string; const endTag: string);
begin
  if (Length(startTag)<>2) or (Length(endTag)<>2) then
    raise Exception.Create('Tags must be of 2 characters');
  fStartTag:=startTag;
  fEndTag:=endTag;
  if fStartTag<>START_TAG then
      startTagChanged:=True
  else
      startTagChanged:=False;
  if fEndTag<>END_TAG then
    endTagChanged:=True
  else
    endTagChanged:=False;
end;
end.
